from dataclasses import dataclass, field
import numpy as np
import pint

from nema.data.data_properties import DataProperties
from nema.utils.units import UNIT_REGISTRY, format_unit_str_for_backend


@dataclass
class FloatVector(DataProperties):
    vector: np.ndarray

    def __post_init__(self):

        # convert to numpy array if needed
        if not isinstance(self.vector, np.ndarray):
            self.vector = np.array(self.vector)

        if len(self.vector.shape) > 1:
            raise ValueError("Vector must be 1D")

    @property
    def data_type(self):
        return "FLOAT_VECTOR.V0"

    def __nema_marshall__(self):
        return {"vector": self.vector.tolist()}

    @classmethod
    def __nema_unmarshall__(cls, data: dict):
        return cls(vector=np.array(data["vector"]))


@dataclass
class FloatVectorWithPhysicalUnits(FloatVector):
    vector: np.ndarray = field(default_factory=list)

    def __post_init__(self):

        # convert to numpy array if needed
        if not isinstance(self.vector, pint.Quantity):
            self.vector = np.array([]) * pint.Unit("dimensionless")

        if len(self.vector.shape) > 1:
            raise ValueError("Vector must be 1D")

    @property
    def data_type(self):
        return "FLOAT_VECTOR_WITH_PHYSICAL_UNITS.V0"

    def __nema_marshall__(self):
        unit_str = format_unit_str_for_backend(self.vector.units)

        return {"vector": self.vector.magnitude.tolist(), "units": unit_str}

    @classmethod
    def __nema_unmarshall__(cls, data: dict):
        vector_magnitude = np.array(data["vector"])
        vector = vector_magnitude * UNIT_REGISTRY(data["units"])
        return cls(vector=vector)


@dataclass
class FloatMatrix(DataProperties):
    matrix: np.ndarray

    def __post_init__(self):

        # convert to numpy array if needed
        if not isinstance(self.matrix, np.ndarray):
            self.matrix = np.array(self.matrix)

        if len(self.matrix.shape) != 2:
            raise ValueError("Matrix must be 2D")

    @property
    def data_type(self):
        return "FLOAT_MATRIX.V0"

    def __nema_marshall__(self):
        return {"matrix": self.matrix.tolist()}

    @classmethod
    def __nema_unmarshall__(cls, data: dict):
        return cls(matrix=np.array(data["matrix"]))



from dataclasses import dataclass, field
import ast
from typing import List, Optional, Set, Any

from nema.data.data_type import DataType

### THIS IS PRETTY MUCH ALL COPIED FROM PROD CODE ###


@dataclass
class DataCoreSubTypeForFunctionArgument:
    type: Optional[DataType] = None
    union: Optional[Set[DataType]] = None
    any: Optional[List[Any]] = None

    def marshall(self):
        return {
            "type": self.type.value if self.type else None,
            "union": [t.value for t in self.union] if self.union else None,
            "any": self.any,
        }


@dataclass
class DataSubTypeForFunctionArgument(DataCoreSubTypeForFunctionArgument):
    list: Optional[DataCoreSubTypeForFunctionArgument] = None
    dictionary: Optional[DataCoreSubTypeForFunctionArgument] = None

    def marshall(self):
        core = super().marshall()
        return {
            **core,
            "list": self.list.marshall() if self.list else None,
            "dictionary": self.dictionary.marshall() if self.dictionary else None,
        }


@dataclass
class FunctionDataArgument:
    id_in_function: str
    description: str
    artifact_type: DataSubTypeForFunctionArgument

    def marshall(self):

        return {
            "id_in_function": self.id_in_function,
            "description": self.description,
            "artifact_type": self.artifact_type.marshall(),
        }


@dataclass
class AppIO:
    """
    Class used to keep the extracted input and output arguments from an app.
    """

    input_data: list[FunctionDataArgument] = field(default_factory=list)
    output_data: list[FunctionDataArgument] = field(default_factory=list)


NEMA_PY_TYPE_MAPPING = {
    "StringValue": DataType.STRING,
    "IntegerValue": DataType.INT,
    "FloatValue": DataType.FLOAT,
    "CurrencyValue": DataType.CURRENCY,
    "FloatValueWithArbitraryUnit": DataType.FLOAT_WITH_ARBITRARY_UNIT_V0,
    "IntValueWithArbitraryUnit": DataType.INT_WITH_ARBITRARY_UNIT_V0,
    "FloatValueWithPhysicalUnit": DataType.FLOAT_WITH_PHYSICAL_UNIT_V0,
    "IntValueWithPhysicalUnit": DataType.INT_WITH_PHYSICAL_UNIT_V0,
    "ArbitraryFile": DataType.ARBITRARY_FILE_V0,
    "ArbitraryFileCollection": DataType.ARBITRARY_FILE_COLLECTION_V0,
    "Dictionary": DataType.DICTIONARY_V0,
    "CSVData": DataType.CSV_V0,
    "Image": DataType.IMAGE_V0,
    "FloatVector": DataType.FLOAT_VECTOR_V0,
    "FloatVectorWithPhysicalUnits": DataType.FLOAT_VECTOR_WITH_PHYSICAL_UNITS_V0,
    "FloatMatrix": DataType.FLOAT_MATRIX_V0,
    "NormalDistribution": DataType.NORMAL_DISTRIBUTION_V0,
    "UniformDistribution": DataType.UNIFORM_DISTRIBUTION_V0,
    "ExponentialDistribution": DataType.EXPONENTIAL_DISTRIBUTION_V0,
    "TriangularDistribution": DataType.TRIANGULAR_DISTRIBUTION_V0,
}


def extract_input_and_output_from_python_contents(code: str) -> AppIO:

    tree = ast.parse(code)
    data_inputs = []
    data_outputs = []

    # Walk through the AST to find the run function
    for node in tree.body:
        if isinstance(node, ast.FunctionDef) and node.name == "run":
            # Extract the argument type annotation of the `run` function
            for arg in node.args.args:
                if arg.annotation and isinstance(arg.annotation, ast.Name):
                    inputs_class_name = arg.annotation.id
                    data_inputs = extract_field_from_ast_tree(tree, inputs_class_name)

            # Extract the return type annotation of the `run` function
            if node.returns and isinstance(node.returns, ast.Name):
                outputs_class_name = node.returns.id
                data_outputs = extract_field_from_ast_tree(tree, outputs_class_name)

    return AppIO(
        input_data=data_inputs,
        output_data=data_outputs,
    )


def extract_field_from_ast_tree(
    tree: ast.Module, class_name: str
) -> list[FunctionDataArgument]:
    fields = []

    # Look for the class definition of the given class_name
    for node in tree.body:
        if isinstance(node, ast.ClassDef) and node.name == class_name:
            # Extract fields from the dataclass
            for stmt in node.body:
                if isinstance(stmt, ast.AnnAssign):
                    key = stmt.target.id

                    if isinstance(stmt.annotation, ast.Subscript) and isinstance(
                        stmt.annotation.value, ast.Name
                    ):
                        if stmt.annotation.value.id == "list":
                            # Extract the type(s) of the list
                            element_types = extract_types(stmt.annotation.slice)

                            artifact_type = DataSubTypeForFunctionArgument(
                                list=element_types,
                            )

                        elif stmt.annotation.value.id == "dict":

                            key_value_types = stmt.annotation.slice
                            if (
                                isinstance(key_value_types, ast.Tuple)
                                and len(key_value_types.elts) == 2
                            ):
                                # Extract key and value types, handling Union in value
                                key_type = extract_types(key_value_types.elts[0])[
                                    0
                                ]  # Keys are usually single types

                                if key_type != "str":
                                    raise ValueError(
                                        f"Invalid key type for dict: {key_type}. Only string is supported."
                                    )

                                value_types = extract_types(key_value_types.elts[1])
                                artifact_type = DataSubTypeForFunctionArgument(
                                    dictionary=value_types,
                                )
                            else:
                                raise ValueError(
                                    f"Invalid dict type annotation: {stmt.annotation}"
                                )

                        elif stmt.annotation.value.id == "Union":
                            # Extract all types within Union
                            if isinstance(stmt.annotation.slice, ast.Tuple):
                                artifact_type = DataSubTypeForFunctionArgument(
                                    union=[
                                        NEMA_PY_TYPE_MAPPING[elt.id]
                                        for elt in stmt.annotation.slice.elts
                                        if isinstance(elt, ast.Name)
                                    ]
                                )
                            elif isinstance(stmt.annotation.slice, ast.Name):
                                artifact_type = DataSubTypeForFunctionArgument(
                                    type=NEMA_PY_TYPE_MAPPING[stmt.annotation.slice.id]
                                )

                        else:
                            artifact_type = DataSubTypeForFunctionArgument(any=[])

                    elif isinstance(stmt.annotation, ast.Name):

                        if stmt.annotation.id.lower() == "any":
                            artifact_type = DataSubTypeForFunctionArgument(any=[])
                        else:
                            artifact_type = DataSubTypeForFunctionArgument(
                                type=NEMA_PY_TYPE_MAPPING[stmt.annotation.id]
                            )

                    else:
                        artifact_type = DataSubTypeForFunctionArgument(any=[])

                    fields.append(
                        FunctionDataArgument(
                            id_in_function=key,
                            description="",
                            artifact_type=artifact_type,
                        )
                    )

    return fields


def extract_types(annotation):
    """Recursively extract types from annotations."""
    if isinstance(annotation, ast.Name):
        # Single type like `int`, `str`, etc.
        return DataCoreSubTypeForFunctionArgument(
            type=NEMA_PY_TYPE_MAPPING[annotation.id]
        )
    elif isinstance(annotation, ast.Subscript):
        # Handle subscripted types (e.g., Union, List, Dict)
        if isinstance(annotation.value, ast.Name) and annotation.value.id == "Union":
            # Extract all types within Union
            if isinstance(annotation.slice, ast.Tuple):
                return DataCoreSubTypeForFunctionArgument(
                    union=[
                        NEMA_PY_TYPE_MAPPING[elt.id]
                        for elt in annotation.slice.elts
                        if isinstance(elt, ast.Name)
                    ]
                )
            elif isinstance(annotation.slice, ast.Name):
                return DataCoreSubTypeForFunctionArgument(
                    type=NEMA_PY_TYPE_MAPPING[annotation.slice.id]
                )
        elif isinstance(annotation.slice, ast.Name):
            if annotation.slice.id.lower() == "any":
                return DataCoreSubTypeForFunctionArgument(any=[])

            return DataCoreSubTypeForFunctionArgument(
                type=NEMA_PY_TYPE_MAPPING[annotation.slice.id]
            )
        elif isinstance(annotation.slice, ast.Tuple):
            return DataCoreSubTypeForFunctionArgument(
                union=[
                    NEMA_PY_TYPE_MAPPING[elt.id]
                    for elt in annotation.slice.elts
                    if isinstance(elt, ast.Name)
                ]
            )
    elif isinstance(annotation, ast.Tuple):
        # Handle tuple types directly
        return DataCoreSubTypeForFunctionArgument(
            union=[
                NEMA_PY_TYPE_MAPPING[elt.id]
                for elt in annotation.elts
                if isinstance(elt, ast.Name)
            ]
        )

    return DataCoreSubTypeForFunctionArgument(any=[])


######


def find_matches_in_workflow_arguments(
    arguments: List[dict],
    id_in_function: str,
):

    return [x for x in arguments if x["id_in_function"] == id_in_function]


def map_workflow_arguments_to_app_arguments(
    workflow_arguments: List[dict],
    app_arguments: List[FunctionDataArgument],
):

    app_inputs = {}
    for input_data_type in app_arguments:
        id_in_function = input_data_type.id_in_function

        argument_matches = find_matches_in_workflow_arguments(
            workflow_arguments, id_in_function
        )

        if (
            input_data_type.artifact_type.type is not None
            or input_data_type.artifact_type.any is not None
        ):
            assert len(argument_matches) == 1

            global_id_of_match = argument_matches[0]["artifact"]

            app_inputs[id_in_function] = global_id_of_match

        elif input_data_type.artifact_type.dictionary is not None:
            this_dict = {}
            for argument_match in argument_matches:
                global_id_of_match = argument_match["artifact"]
                this_dict[argument_match.member_type.key_in_dictionary] = (
                    global_id_of_match
                )
            app_inputs[id_in_function] = this_dict

        elif input_data_type.artifact_type.list is not None:
            this_list = [None for _ in argument_matches]
            for argument_match in argument_matches:
                global_id_of_match = argument_match["artifact"]
                idx = argument_match["member_type"]["index_in_list"]
                this_list[idx] = global_id_of_match
            app_inputs[id_in_function] = this_list

    return app_inputs



from typing import Callable
from functools import wraps
from dataclasses import dataclass
import uuid
from typing import Dict, Any
from types import ModuleType

from nema.run.utils import (
    convert_app_input_to_dict,
    convert_app_output_to_nema_data,
    add_code_directory_to_sys_path,
)
from nema.workflow.workflow import Workflow


@dataclass
class InternalWorkflowOutput:
    input_data: Any
    output_data: Any
    input_global_id_mapping: Dict[str, int]
    output_global_id_mapping: Dict[str, int]


def workflow(
    input_global_id_mapping: Dict[str, int],
    output_global_id_mapping: Dict[str, int],
    job_id=None,
):
    # create random string if not provided
    actual_job_id = job_id if job_id else str(uuid.uuid4())

    def decorator(func: Callable):

        # Attach the mappings and job ID as attributes to the function
        func.__workflow_attributes__ = {
            "input_global_id_mapping": input_global_id_mapping,
            "output_global_id_mapping": output_global_id_mapping,
            "job_id": actual_job_id,
        }

        @wraps(func)
        def wrapper(*args, **kwargs):

            if len(args) > 0 or len(kwargs) > 0:
                # not running from Nema
                return func(*args, **kwargs)

            # now we need to pull data from Nema and upload the results back

            # check that all global IDs are valid
            for key, g_id in input_global_id_mapping.items():
                if g_id <= 0:
                    raise ValueError(
                        f"Global ID for input {key} is not set. Run `nema-python workflow create-data` to create the data."
                    )

            for key, g_id in output_global_id_mapping.items():
                if g_id <= 0:
                    raise ValueError(
                        f"Global ID for output {key} is not set. Run `nema-python workflow create-data` to create the data."
                    )

            # get the type of func
            annotations = func.__annotations__
            arg_name = list(annotations)[0]
            input_type = annotations[arg_name]
            converted_input = convert_app_input_to_dict(
                input_global_id_mapping, branch="main"
            )

            typed_argument = input_type(**converted_input)

            result = func(typed_argument)

            return InternalWorkflowOutput(
                input_data=typed_argument,
                output_data=result,
                input_global_id_mapping=input_global_id_mapping,
                output_global_id_mapping=output_global_id_mapping,
            )

        return wrapper

    # # close all the data properties in the collection
    # data_collection.close()

    return decorator


def run_workflow(wf: Workflow, python_file: str):

    # to make sure that local imports work, we need to add the directory the code is in to the sys path
    add_code_directory_to_sys_path(python_file)

    # read the python file
    with open(python_file, "r") as f:
        code_to_execute = f.read()

    # execute the application code
    user_module = ModuleType("user_module")

    # load code into module
    exec(code_to_execute, user_module.__dict__)

    # execute code
    output: InternalWorkflowOutput = user_module.run()

    # put output back into the right format
    converted_nema_data = convert_app_output_to_nema_data(
        output.output_data, output.output_global_id_mapping
    )

    # sync output to API
    used_data = []
    for key_used, value_used in output.input_global_id_mapping.items():
        used_data.append(
            {
                "artifact": value_used,
                "id_in_function": key_used,
            }
        )

    wf.process_local_update(
        updated_data=converted_nema_data,
        used_data=used_data,
    )



from dataclasses import dataclass
from typing import Optional, List
import json
import os
import shutil

from nema.connectivity import ConnectivityManager, CONNECTIVITY_CONFIG
from nema.data import Data


def get_all_files_from_output_folder(output_folder: str):

    files = []
    for root, _, found_files in os.walk(output_folder):
        for file_name in found_files:
            file_path = os.path.join(root, file_name)
            files.append(file_path)

    return files


@dataclass
class Workflow:
    global_id: int
    name: str = ""
    description: str = ""
    output_folder: Optional[str] = None

    def marshall_create(self):
        return {
            "name": self.name,
            "description": self.description,
            "workflow_type": "NEMA.EXTERNAL.V0",
            "workflow_properties": {"external_workflow_type": "PYTHON"},
        }

    def create(self):
        conn = ConnectivityManager()
        global_id = conn.create_workflow(self.marshall_create())
        return global_id

    def process_local_update(
        self,
        updated_data: List[Data],
        used_data: List[dict],
    ):
        conn = ConnectivityManager()

        output_folder = self.output_folder or CONNECTIVITY_CONFIG.nema_data_folder

        files_output_folder = os.path.join(output_folder, "data-output")
        shutil.rmtree(files_output_folder, ignore_errors=True)
        os.makedirs(files_output_folder, exist_ok=True)

        raw_data_output = [
            (
                "updated_data",
                json.dumps(
                    data.process_output_data(destination_folder=files_output_folder)
                ),
            )
            for data in updated_data
        ]

        raw_uses_data = [
            (
                "used_data",
                json.dumps(i_data),
            )
            for i_data in used_data
        ]

        raw_data = raw_data_output + raw_uses_data

        # read files from output folder
        files = get_all_files_from_output_folder(files_output_folder)

        conn.push_local_workflow_update(self.global_id, raw_data, files=files)

    def push_outputs_to_API(
        self,
        updated_data: List[Data],
        branch_name: Optional[str] = None,
        local: bool = False,
    ):
        conn = ConnectivityManager()

        output_folder = self.output_folder or CONNECTIVITY_CONFIG.nema_data_folder

        files_output_folder = os.path.join(output_folder, "data-output")
        shutil.rmtree(files_output_folder, ignore_errors=True)
        os.makedirs(files_output_folder, exist_ok=True)

        raw_data = [
            (
                "updated_data",
                json.dumps(
                    data.process_output_data(destination_folder=files_output_folder)
                ),
            )
            for data in updated_data
        ]

        # read files from output folder
        files = get_all_files_from_output_folder(files_output_folder)

        conn.push_workflow_outputs(
            self.global_id, raw_data, branch=branch_name, files=files, local=local
        )

    def get_workflow_details(self) -> dict:
        conn = ConnectivityManager()

        return conn.get_workflow_io(self.global_id)



from .data import Data, FileData

from .data_properties import (
    StringValue,
    IntegerValue,
    FloatValue,
    CurrencyValue,
    FloatValueWithArbitraryUnit,
    IntValueWithArbitraryUnit,
    FloatValueWithPhysicalUnit,
    IntValueWithPhysicalUnit,
    ArbitraryFile,
    ArbitraryFileCollection,
    Dictionary,
)
from .tabular import CSVData
from .plots import Image
from .linear_algebra import FloatVector, FloatMatrix, FloatVectorWithPhysicalUnits
from .distributions import (
    NormalDistribution,
    UniformDistribution,
    ExponentialDistribution,
    TriangularDistribution,
)



import pint

try:
    from pint.delegates.formatter.plain import DefaultFormatter as DefaultPintFormatter

    UNIT_REGISTRY = pint.UnitRegistry()
    DEFAULT_PINT_FORMATTER = DefaultPintFormatter()
    DEFAULT_PINT_FORMATTER._registry = UNIT_REGISTRY

    def format_unit_str_for_backend(units: pint.Unit):
        return (
            DEFAULT_PINT_FORMATTER.format_unit(units, "~")
            .replace(" ", "")
            .replace("**", "^")
        )

except ImportError:
    UNIT_REGISTRY = pint.UnitRegistry()
    DEFAULT_PINT_FORMATTER = None

    def format_unit_str_for_backend(units: pint.Unit):
        return format(units, "~").replace(" ", "").replace("**", "^")



from typing import List
import os
import sys

from nema.data.data_properties import DataProperties
from nema.connectivity import CONNECTIVITY_CONFIG
from nema.data.data import Data, FileData


def get_data_properties_from_global_id(global_id: int, branch: str) -> DataProperties:
    data = Data.init_from_cloud_and_download(global_id, branch=branch)

    if data.data.is_blob_data:
        data = FileData(
            _global_id=global_id,
            _data=data.data,
            _input_folder=CONNECTIVITY_CONFIG.nema_data_folder,
            _branch=branch,
        )

        # TODO: this is super messy!
        if hasattr(data.data, "get_contents"):

            data.contents  # this downloads the file and extracts the contents

        elif hasattr(data.data, "get_file_name"):

            data.get_file_path()  # this downloads the file

    return data.data  # return the data properties


def convert_app_input_to_dict(inputs: dict, branch: str = "main") -> dict:
    """The input is a dictionary of global IDs. This function will convert the global IDs to the actual data properties"""
    converted_input = {}
    for key, value in inputs.items():
        if isinstance(value, list):
            converted_input[key] = [
                get_data_properties_from_global_id(v, branch) for v in value
            ]
        elif isinstance(value, dict):
            converted_input[key] = convert_app_input_to_dict(value, branch)
        else:
            converted_input[key] = get_data_properties_from_global_id(value, branch)

    return converted_input


def convert_app_output_to_nema_data(output, output_dict: dict) -> List[Data]:
    """The output is a dictionary of data properties. This function will convert the properties to Nema Data objects, which can be uploaded to Nema"""
    converted_output = []
    for key, value in output_dict.items():
        this_output = getattr(output, key)
        if isinstance(value, list):
            for idx, lv in enumerate(value):
                converted_output.append(
                    Data.init_from_properties(
                        global_id=lv, data_properties=this_output[idx]
                    )
                )
        elif isinstance(value, dict):
            for lk, lv in zip(value, getattr(output, key)):
                converted_output.append(
                    Data.init_from_properties(
                        global_id=lv, data_properties=this_output[lk]
                    )
                )
        else:
            converted_output.append(
                Data.init_from_properties(value, this_output, id_in_function=key)
            )

    return converted_output


def add_code_directory_to_sys_path(file_path):
    """Add the file's directory to sys.path."""
    code_dir = os.path.dirname(os.path.abspath(file_path))
    if code_dir not in sys.path:
        sys.path.insert(0, code_dir)



from dataclasses import dataclass

from nema.data.data_properties import DataProperties


@dataclass
class NormalDistribution(DataProperties):
    mean: float
    std_dev: float

    @property
    def data_type(self):
        return "NORMAL_DISTRIBUTION.V0"


@dataclass
class UniformDistribution(DataProperties):
    lower_bound: float
    upper_bound: float

    @property
    def data_type(self):
        return "UNIFORM_DISTRIBUTION.V0"


@dataclass
class ExponentialDistribution(DataProperties):
    rate: float

    @property
    def data_type(self):
        return "EXPONENTIAL_DISTRIBUTION.V0"


@dataclass
class TriangularDistribution(DataProperties):
    lower_bound: float
    upper_bound: float
    mode: float

    @property
    def data_type(self):
        return "TRIANGULAR_DISTRIBUTION.V0"



from dataclasses import dataclass
from typing import Optional
import json
from typing import List

from nema.connectivity import ConnectivityManager

from .utils import extract_input_and_output_from_python_contents, AppIO


@dataclass
class App:
    global_id: int
    name: str = ""
    description: str = ""
    output_folder: Optional[str] = None
    code: Optional[str] = None
    app_io: Optional[AppIO] = None

    def download_code(self, branch: str) -> str:
        conn = ConnectivityManager()

        return conn.retrieve_app_code(self.global_id, branch=branch)

    def extract_io_from_code(self) -> None:
        self.app_io = extract_input_and_output_from_python_contents(self.code)

    def to_raw_data(self) -> List[tuple]:

        data = [
            ("name", self.name),
            ("description", self.description),
            ("function_type", "NEMA.APP_EXTERNAL.V0"),
            ("function_properties", json.dumps({"external_function_type": "PYTHON"})),
        ]

        if self.code:

            self.extract_io_from_code()

            for in_arg in self.app_io.input_data:
                data.append(("input_data_arguments", json.dumps(in_arg.marshall())))

            for out_arg in self.app_io.output_data:
                data.append(("output_data_arguments", json.dumps(out_arg.marshall())))

        return data

    def create(self) -> int:
        conn = ConnectivityManager()

        raw_data = self.to_raw_data()

        global_id = conn.create_app(raw_data)
        return global_id

    def update(self) -> None:
        conn = ConnectivityManager()

        raw_data = self.to_raw_data()

        conn.update_app(raw_data, global_id=self.global_id)

    def get_all_workflows(self) -> List:
        conn = ConnectivityManager()

        return conn.get_workflows_for_app(self.global_id)



from typing import List, Type
import json
import re

from nema.data.data import Data
from nema.connectivity import ConnectivityManager


def get_new_data_for_workflow(
    global_id_mapping: dict, argument_type: Type
) -> List[Data]:

    data_to_create: List[Data] = []

    # do this for outputs to TODO
    for key, global_id in global_id_mapping.items():
        if global_id > 0:
            print(f"Data for key '{key}' already exists. Skipping.")
            continue

        # find class for properties
        data_properties_cls = argument_type.__annotations__[key]
        new_data_properties = data_properties_cls()

        data = Data.init_from_properties(
            global_id=0,
            data_properties=new_data_properties,
            id_in_function=key,
        )
        data_to_create.append(data)

    return data_to_create


def submit_new_data_to_nema(data: List[Data]):

    raw_data = []
    for idx, d in enumerate(data):

        raw_data += [
            (
                f"entries[{idx}].data_properties",
                json.dumps(d.data.__nema_marshall__()),
            ),
            (
                f"entries[{idx}].description",
                "<p></p>",
            ),
            (
                f"entries[{idx}].name",
                d.id_in_function,
            ),
            (
                f"entries[{idx}].data_type",
                d.data.data_type,
            ),
        ]

    connectivity_manager = ConnectivityManager()

    resulting_data_ids = connectivity_manager.push_batch_data(raw_data)

    for idx, data_id in enumerate(resulting_data_ids):
        data[idx].global_id = data_id


def update_workflow_file_with_new_global_ids(
    code_str: str, new_data_inputs: List[Data], new_data_outputs: List[Data]
):
    input_mapping_dict = {d.id_in_function: d.global_id for d in new_data_inputs}
    output_mapping_dict = {d.id_in_function: d.global_id for d in new_data_outputs}

    def update_mapping(mapping_block, mapping_dict):
        for key, new_id in mapping_dict.items():
            # Replace the number after "key":
            mapping_block = re.sub(
                rf'("{key}"\s*:\s*)\d+',  # Match the key and the number after ":"
                lambda match: f"{match.group(1)}{new_id}",  # Replace with the new number
                mapping_block,
            )
        return mapping_block

    # Replace input_global_id_mapping block
    input_match = re.search(r"input_global_id_mapping=\{(.*?)\}", code_str, re.DOTALL)
    if input_match:
        input_block = input_match.group(1)
        updated_input_block = update_mapping(input_block, input_mapping_dict)
        code_str = code_str.replace(input_block, updated_input_block)

    # Replace output_global_id_mapping block
    output_match = re.search(r"output_global_id_mapping=\{(.*?)\}", code_str, re.DOTALL)
    if output_match:
        output_block = output_match.group(1)
        updated_output_block = update_mapping(output_block, output_mapping_dict)
        code_str = code_str.replace(output_block, updated_output_block)

    return code_str



import click
from types import ModuleType
from typing import List
import requests
from typing import Optional

from nema.utils.global_config import (
    GLOBAL_CONFIG,
    GlobalConfig,
    GlobalConfigWorkflow,
    GlobalConfigFunction,
)
from nema.app import App
from nema.workflow.workflow import Workflow
from nema.connectivity.connectivity_manager import save_auth_data
from nema.run.workflow import run_workflow
from nema.run.app import run_wf_from_app
from nema.run.utils import add_code_directory_to_sys_path
from nema.utils.create_data_for_workflow import (
    get_new_data_for_workflow,
    submit_new_data_to_nema,
    update_workflow_file_with_new_global_ids,
)


def authenticate_user(username, password):

    # Authenticate the user
    if GLOBAL_CONFIG.is_set:

        tenant_url = GLOBAL_CONFIG.tenant_api_url

        login_url = f"{tenant_url}/authentication/login"

    else:
        raise Exception(
            "There is no nema.toml file in this directory. Please run `nema init` to create one."
        )

    # Make a request to the login URL
    response = requests.post(
        login_url, json={"username": username, "password": password}
    )

    if not response.ok:
        if response.status_code == 401:
            print("Invalid credentials")
        else:
            print(response.status_code, response.text)
            print("Failed to login")
        return None

    # save the authentication and refresh token to the home directory
    response_data = response.json()
    tokens = response_data["tokens"]
    refresh_token = tokens["refresh_token"]
    access_token = tokens["access_token"]

    # create the directory if it does not exist
    save_auth_data(refresh_token=refresh_token, access_token=access_token)

    print("Login successful \U00002705")


@click.group()
def cli():
    pass


@cli.command()
def login():
    """Login to Nema."""
    username = click.prompt("Please enter your username or email address")
    password = click.prompt("Please enter your password", hide_input=True)
    authenticate_user(username, password)


@cli.command()
def init():
    "Initialize nema.toml file"
    print("Initializing nema.toml file")

    new_global_config = GlobalConfig()

    new_global_config.project_url = click.prompt(
        "Please enter the project URL", type=str
    )

    new_artifact_type = click.prompt(
        "Please enter whether you want to create a workflow or a function",
        type=click.Choice(["workflow", "function"]),
    )
    new_identifier = click.prompt(
        f"Please enter a {new_artifact_type} identifier (this is only used locally)",
        type=str,
        default=f"my-first-{new_artifact_type}",
    )
    new_name = click.prompt(
        f"Please enter a name for the {new_artifact_type}",
        type=str,
        default=new_identifier,
    )
    new_description = click.prompt(
        f"Please enter a {new_artifact_type} description",
        type=str,
        default=f"A Python {new_artifact_type}",
    )

    if new_artifact_type == "workflow":
        new_workflow = GlobalConfigWorkflow(
            key=new_identifier,
            name=new_name,
            description=new_description,
            script=f"nema_{new_artifact_type}.py",
        )

        new_global_config.workflows[new_identifier] = new_workflow
    else:
        new_function = GlobalConfigFunction(
            key=new_identifier,
            name=new_name,
            description=new_description,
            script=f"nema_{new_artifact_type}.py",
        )

        new_global_config.functions[new_identifier] = new_function

    new_global_config.save()


@cli.group()
def workflow():
    pass


@workflow.command()
@click.argument("identifier", required=False)
def init(identifier: Optional[str]):
    "Record the workflow in Nema"

    if identifier is None:
        print("No workflow specified -- initializing all workflows")
        all_identifiers = GLOBAL_CONFIG.workflows.keys()
    else:
        all_identifiers = [identifier]

    for this_identifier in all_identifiers:
        print(f"Initializing workflow with identifier '{this_identifier}'")
        existing_workflow = GLOBAL_CONFIG.workflows[this_identifier]

        if existing_workflow.global_id > 0:
            print("Workflow already exists. Skipping.")
            continue

        workflow = Workflow(
            global_id=0,
            name=existing_workflow.name,
            description=existing_workflow.description,
        )

        global_id = workflow.create()
        print(f"Workflow successfully created with global id {global_id}. \U00002705")

        existing_workflow.global_id = global_id

    # save the global ID to the config file
    GLOBAL_CONFIG.save()


@workflow.command()
@click.argument("identifier", required=False)
def run(identifier: Optional[str]):
    "Run the workflow"

    if identifier is None:
        print("No workflow specified -- running all workflows")
        all_identifiers = GLOBAL_CONFIG.workflows.keys()
    else:
        all_identifiers = [identifier]

    for this_identifier in all_identifiers:
        print(f"Running workflow with identifier '{this_identifier}'")

        existing_workflow = GLOBAL_CONFIG.workflows[this_identifier]

        python_file = existing_workflow.script

        this_wf = Workflow(
            global_id=existing_workflow.global_id,
            name=existing_workflow.name,
            description=existing_workflow.description,
            # output_folder=output_folder,
        )

        run_workflow(this_wf, python_file)

        print(
            f"Workflow '{this_identifier}' successfully run \U0001F680 and results uploaded to Nema. \U00002705"
        )


@workflow.command(name="create-data")
@click.argument("identifier", required=False)
def create_data(identifier: Optional[str]):
    "Create data for the workflow"

    if identifier is None:
        print("No workflow specified -- creating data for all workflows")
        all_identifiers = GLOBAL_CONFIG.workflows.keys()
    else:
        all_identifiers = [identifier]

    for this_identifier in all_identifiers:
        print(f"Creating data for workflow with identifier '{this_identifier}'")

        existing_workflow = GLOBAL_CONFIG.workflows[this_identifier]

        python_file = existing_workflow.script

        # read the python file
        with open(python_file, "r") as f:
            code_to_execute = f.read()

        # to make sure that local imports work, we need to add the directory the code is in to the sys path
        add_code_directory_to_sys_path(python_file)

        # execute the application code
        user_module = ModuleType("user_module")

        # load code into module
        exec(code_to_execute, user_module.__dict__)

        # execute code
        run_function = user_module.run

        annotations = run_function.__annotations__
        arg_name = list(annotations)[0]
        input_type = annotations[arg_name]
        output_type = annotations["return"]

        data_to_create_inputs = get_new_data_for_workflow(
            run_function.__workflow_attributes__["input_global_id_mapping"], input_type
        )
        data_to_create_outputs = get_new_data_for_workflow(
            run_function.__workflow_attributes__["output_global_id_mapping"],
            output_type,
        )

        data_to_create = data_to_create_inputs + data_to_create_outputs

        # create the data
        submit_new_data_to_nema(data_to_create)

        new_workflow_file = update_workflow_file_with_new_global_ids(
            code_to_execute, data_to_create_inputs, data_to_create_outputs
        )

        # save the new workflow file
        with open(python_file, "w") as f:
            f.write(new_workflow_file)


@cli.group()
def function():
    pass


@function.command()
@click.argument("identifier", required=False)
def init(identifier: Optional[str]):
    "Record the function in Nema"

    if identifier is None:
        print("No function specified -- initializing all functions")
        all_identifiers = GLOBAL_CONFIG.functions.keys()
    else:
        all_identifiers = [identifier]

    for this_identifier in all_identifiers:
        print(f"Initializing function with identifier '{this_identifier}'")
        existing_function = GLOBAL_CONFIG.functions[this_identifier]

        if existing_function.global_id > 0:
            print("Function already exists. Skipping.")
            continue

        python_file = existing_function.script

        # read the python file
        try:
            with open(python_file, "r") as f:
                code_to_execute = f.read()
        except FileNotFoundError:
            print(f"Python file {python_file} not found.")
            code_to_execute = None

        function = App(
            global_id=0,
            name=existing_function.name,
            description=existing_function.description,
            code=code_to_execute,
        )

        global_id = function.create()
        print(f"Function successfully created with global id {global_id}. \U00002705")

        existing_function.global_id = global_id

    # save the global ID to the config file
    GLOBAL_CONFIG.save()


@function.command()
@click.argument("identifier", required=False)
def update(identifier: Optional[str]):
    "Update the function name, description and arguments in Nema" ""

    if identifier is None:
        print("No function specified -- updating all functions")
        all_identifiers = GLOBAL_CONFIG.functions.keys()
    else:
        all_identifiers = [identifier]

    for this_identifier in all_identifiers:
        print(f"Updating function with identifier '{this_identifier}'")
        existing_function = GLOBAL_CONFIG.functions[this_identifier]

        if existing_function.global_id == 0:
            print("Function needs to be initialized with Nema first. Skipping.")
            continue

        python_file = existing_function.script

        # read the python file
        with open(python_file, "r") as f:
            code_to_execute = f.read()

        function = App(
            global_id=existing_function.global_id,
            name=existing_function.name,
            description=existing_function.description,
            code=code_to_execute,
        )

        function.update()
        print(
            f"Function '{this_identifier}' successfully updated \U0001F680. Note that the associated workflows are not yet run."
        )


@function.command()
@click.argument("identifier", required=False)
@click.option(
    "--only-outdated",
    is_flag=True,
    default=False,
    help="Only run the outdated workflows",
)
def run(identifier: Optional[str], only_outdated: bool = False):
    "Run the function"

    if identifier is None:
        print("No function specified -- running all workflows\n")
        all_identifiers = GLOBAL_CONFIG.functions.keys()
    else:
        all_identifiers = [identifier]

    for this_identifier in all_identifiers:
        print(f"Running function with identifier '{this_identifier}'")

        existing_function = GLOBAL_CONFIG.functions[this_identifier]

        python_file = existing_function.script

        # read the python file
        with open(python_file, "r") as f:
            code_to_execute = f.read()

        # to make sure that local imports work, we need to add the directory the code is in to the sys path
        add_code_directory_to_sys_path(python_file)

        function = App(
            global_id=existing_function.global_id,
            name=existing_function.name,
            description=existing_function.description,
            code=code_to_execute,
        )

        function.extract_io_from_code()

        all_workflows_that_use_this_app = function.get_all_workflows()

        if len(all_workflows_that_use_this_app) == 0:
            print("No workflows use this function. Skipping.")
            continue

        for this_workflow_details in all_workflows_that_use_this_app:

            wf_global_id = this_workflow_details["global_id"]
            wf_name = this_workflow_details["name"]

            print(f"Running workflow '{wf_name}' (#{wf_global_id})..")

            this_wf = Workflow(
                global_id=wf_global_id,
                name=this_workflow_details["name"],
                description=this_workflow_details["description"],
            )

            has_run = run_wf_from_app(
                this_wf, function, only_run_outdated=only_outdated
            )

            if has_run:
                print(
                    f"Workflow '{wf_name}' (#{wf_global_id}) successfully run \U0001F680 and results uploaded to Nema. \U00002705 \n"
                )


if __name__ == "__main__":
    cli()



from dataclasses import dataclass

from nema.data.data_properties import DataProperties, FileDataProperties


@dataclass
class FigureDataProperties(DataProperties):
    pass


@dataclass
class Image(FigureDataProperties, FileDataProperties):

    def __post_init__(self):
        self.extension = self.extension or "png"
        return super().__post_init__()

    @property
    def data_type(self):
        return "IMAGE.V0"



from types import ModuleType

from nema.workflow.workflow import Workflow
from nema.app.app import App
from nema.app.utils import map_workflow_arguments_to_app_arguments
from nema.run.utils import convert_app_output_to_nema_data, convert_app_input_to_dict


def run_wf_from_app(wf: Workflow, app: App, only_run_outdated: bool = False) -> bool:

    branch_name = "main"

    wf_details = wf.get_workflow_details()

    if only_run_outdated:
        local_status = wf_details["local_status"]["status"]
        if local_status == "UP_TO_DATE":
            print("Skipping workflow as it is up to date\n")
            return False

    # execute the application code
    user_module = ModuleType("user_module")

    # retrieve inputs for this workflow
    formatted_input_data = map_workflow_arguments_to_app_arguments(
        wf_details["input_data"], app.app_io.input_data
    )
    formatted_output_data = map_workflow_arguments_to_app_arguments(
        wf_details["output_data"], app.app_io.output_data
    )

    converted_input = convert_app_input_to_dict(
        formatted_input_data, branch=branch_name
    )

    # load code into module
    exec(app.code, user_module.__dict__)

    inputs = user_module.Inputs(**converted_input)

    # execute code
    output = user_module.run(inputs)

    # put output back into the right format
    converted_nema_data = convert_app_output_to_nema_data(output, formatted_output_data)

    # sync output to API
    wf.push_outputs_to_API(converted_nema_data, branch_name=branch_name, local=True)

    return True
